# -*- coding: utf-8 -*-
"""
/***************************************************************************
 IndoorGMLReader
                                 A QGIS plugin
 This plugin reads the content of an IndoorGML schema file.
                              -------------------
        begin                : 2016-11-06
        git sha              : $Format:%H$
        copyright            : (C) 2016 by Alfonso Mariscal Garcia
        email                : amariscalg@uoc.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
# Import qgis libraries in order to interoperate with QGIS elements.
import qgis
from qgis.core import *
from qgis.gui import *
from qgis.utils import iface
# Import PyQt4 libraries in order to satisfy functional needs. (QMessageBox, QFileDialog)
from PyQt4 import QtCore, QtGui
from PyQt4.QtCore import QSettings, QTranslator, qVersion, QCoreApplication
from PyQt4.QtGui import QAction, QIcon, QFileDialog, QMessageBox
# Import libraries in order to explore xml file.
from xml.dom import minidom
import sys
# Initialize Qt resources from file resources.py
import resources
# Import the code for the dialog
from indoorgml_reader_dialog import IndoorGMLReaderDialog
import os.path

class IndoorGMLReader:
    """QGIS Plugin Implementation."""

    def __init__(self, iface):
        """Constructor.

        :param iface: An interface instance that will be passed to this class
            which provides the hook by which you can manipulate the QGIS
            application at run time.
        :type iface: QgsInterface
        """
        
        # Create the dialog (after translation) and keep reference
        self.dlg = IndoorGMLReaderDialog()
        # Save reference to the QGIS interface
        self.iface = iface
        # initialize plugin directory
        self.plugin_dir = os.path.dirname(__file__)
        # initialize locale
        locale = QSettings().value('locale/userLocale')[0:2]
        locale_path = os.path.join(
            self.plugin_dir,
            'i18n',
            'IndoorGMLReader_{}.qm'.format(locale))

        if os.path.exists(locale_path):
            self.translator = QTranslator()
            self.translator.load(locale_path)

            if qVersion() > '4.3.3':
                QCoreApplication.installTranslator(self.translator)


        # Declare instance attributes
        self.actions = []
        self.menu = self.tr(u'&IndoorGML Reader')
        # TODO: We are going to let the user set this up in a future iteration
        self.toolbar = self.iface.addToolBar(u'IndoorGMLReader')
        self.toolbar.setObjectName(u'IndoorGMLReader')
        # Initialize load button and field.
        self.dlg.gmlLine.clear()
        self.dlg.loadFile.clicked.connect(self.load_indoorGML_file)
          
          
    # noinspection PyMethodMayBeStatic
    def tr(self, message):
        """Get the translation for a string using Qt translation API.

        We implement this ourselves since we do not inherit QObject.

        :param message: String for translation.
        :type message: str, QString

        :returns: Translated version of message.
        :rtype: QString
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate('IndoorGMLReader', message)


    def add_action(
        self,
        icon_path,
        text,
        callback,
        enabled_flag=True,
        add_to_menu=True,
        add_to_toolbar=True,
        status_tip=None,
        whats_this=None,
        parent=None):
        """Add a toolbar icon to the toolbar.

        :param icon_path: Path to the icon for this action. Can be a resource
            path (e.g. ':/plugins/foo/bar.png') or a normal file system path.
        :type icon_path: str

        :param text: Text that should be shown in menu items for this action.
        :type text: str

        :param callback: Function to be called when the action is triggered.
        :type callback: function

        :param enabled_flag: A flag indicating if the action should be enabled
            by default. Defaults to True.
        :type enabled_flag: bool

        :param add_to_menu: Flag indicating whether the action should also
            be added to the menu. Defaults to True.
        :type add_to_menu: bool

        :param add_to_toolbar: Flag indicating whether the action should also
            be added to the toolbar. Defaults to True.
        :type add_to_toolbar: bool

        :param status_tip: Optional text to show in a popup when mouse pointer
            hovers over the action.
        :type status_tip: str

        :param parent: Parent widget for the new action. Defaults None.
        :type parent: QWidget

        :param whats_this: Optional text to show in the status bar when the
            mouse pointer hovers over the action.

        :returns: The action that was created. Note that the action is also
            added to self.actions list.
        :rtype: QAction
        """

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            self.toolbar.addAction(action)

        if add_to_menu:
            self.iface.addPluginToMenu(
                self.menu,
                action)

        self.actions.append(action)

        return action

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        icon_path = ':/plugins/IndoorGMLReader/icon.png'
        self.add_action(
            icon_path,
            text=self.tr(u'IndoorGML Reader'),
            callback=self.run,
            parent=self.iface.mainWindow())


    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""
        for action in self.actions:
            self.iface.removePluginMenu(
                self.tr(u'&IndoorGML Reader'),
                action)
            self.iface.removeToolBarIcon(action)
        # remove the toolbar
        del self.toolbar

    # Load the IndoorGML File to the plugin.
    def load_indoorGML_file(self):
        indoorGMLFile = QFileDialog.getOpenFileName(self.dlg, "Load IndoorGML file ","", '*.gml')
        
        # Once the file is defined, we upload the information to a variable.        
        if indoorGMLFile:
            self.dlg.gmlLine.setText(indoorGMLFile)
            file = self.dlg.gmlLine.displayText()
            gml = minidom.parse(file)
            # Once the file is uploaded, the exploration tree and the layers are also created.
            self.printCellSpaces(gml)
            self.printStates(gml)
            self.printTransitions(gml)
            self.create_tree(gml)
        else:
            self.showMessage()

    
    # Create the exploration tree of the IndoorGML file.
    def create_tree(self, gml):
        tree = self.dlg.tree
        header = QtGui.QTreeWidgetItem(["Features Information","Value"])
        tree.setHeaderItem(header)
        tree.header().setResizeMode(QtGui.QHeaderView.ResizeToContents)
        tree.header().setStretchLastSection(False)
        self.dlg.tree.clear()
        
        # In order to launch a trigger when we click on a item.
        self.dlg.connect (self.dlg.tree, QtCore.SIGNAL ("itemClicked(QTreeWidgetItem*, int)"), self.displayXlinkInformation)
        
        # We set "IndoorGML" tag name as the root of the exploration tree.
        root = QtGui.QTreeWidgetItem(tree, ["IndoorGML"])
        IFs = gml.getElementsByTagName('IndoorFeatures')
        if IFs.length >= 1:
            var = QtGui.QTreeWidgetItem(root, ['Properties'])
            self.getAttributes(IFs[0],var)
        
        # The IndoorGML document will be set under a PrimalSpaceFeatures item.
        PSF = gml.getElementsByTagName('PrimalSpaceFeatures')
        
        # The seconday main tree of the exploration tree will be the MultiLayeredGraph
        MLG = gml.getElementsByTagName('MultiLayeredGraph')
    
        if PSF.length >= 1:      
            for psf in PSF:
                # We go through the whole of "PrimalSpaceFeatures" items, and display their attributes.
                A = QtGui.QTreeWidgetItem(root, ['PrimalSpaceFeatures'])
                varA = QtGui.QTreeWidgetItem(A, ['Properties'])
                self.getAttributes(psf,varA)
                # Now we search for "CellSpace" elements inside of this PrimalSpace.
                CS = psf.getElementsByTagName('CellSpace')
                for cs in CS:
                    varA = QtGui.QTreeWidgetItem(A, ['CellSpace'])
                    varAA = QtGui.QTreeWidgetItem(varA, ['Properties'])
                    self.getAttributes(cs,varAA)
                    varAB = QtGui.QTreeWidgetItem(varA, ['Geometry'])
                    # Exploring child nodes and their attributes.
                    childs = cs.childNodes
                    for child in childs:
                        self.getAttributes(child,varAB)
                        self.getNodes(child.childNodes,400,varAB)
                    # We define a new identation to polygon elements.
                    polygons = cs.getElementsByTagName('gml:Polygon')
                    for polygon in polygons:
                        varABA = QtGui.QTreeWidgetItem(varAB, [polygon.nodeName])
                        self.getAttributes(polygon,varABA)
                        self.getNodes(polygon.childNodes,400,varABA)
           
           
        if MLG.length >= 1:   
            for mlg in MLG:
                B = QtGui.QTreeWidgetItem(root, ['MultiLayeredGraph'])
                varB = QtGui.QTreeWidgetItem(B, ['Properties'])
                self.getAttributes(mlg,varB)
                # Identifying spaceLayers within the MultiLayeredGraph
                SL = mlg.getElementsByTagName('SpaceLayer')
                for sl in SL:
                    varB = QtGui.QTreeWidgetItem(B, ['SpaceLayer'])
                    varBA = QtGui.QTreeWidgetItem(varB, ['Properties'])
                    self.getAttributes(sl,varBA)
                    # Getting nodes and edges of each spaceLayer
                    N = sl.getElementsByTagName('nodes')
                    E = sl.getElementsByTagName('edges')
                    for n in N:
                        varBA = QtGui.QTreeWidgetItem(varB, ['Node'])
                        varBAA = QtGui.QTreeWidgetItem(varBA,['Properties'])
                        self.getAttributes(n,varBAA)
                        # Getting the States of each Node
                        S = n.getElementsByTagName('State')
                        for s in S:
                            varBAA = QtGui.QTreeWidgetItem(varBA, ['State'])
                            self.getAttributes(s,varBAA)
                            self.getNodes(s.childNodes,400,varBAA)   
                            points = s.getElementsByTagName('gml:Point')
                            for point in points:
                                varBAAA = QtGui.QTreeWidgetItem(varBAA, [point.nodeName])
                                self.getAttributes(point,varBAAA)
                                self.getNodes(point.childNodes,400,varBAAA)
                    for e in E:
                        varBB = QtGui.QTreeWidgetItem(varB, ['Edge'])
                        varBBA = QtGui.QTreeWidgetItem(varBB,['Properties'])
                        self.getAttributes(n,varBBA)
                        # Getting the Transitions of each edge
                        T = e.getElementsByTagName('Transition')
                        for t in T:
                            varBBB = QtGui.QTreeWidgetItem(varBB, ['Transition'])
                            self.getAttributes(t,varBBB)
                            self.getNodes(t.childNodes,400,varBBB)
                            lineStrings = t.getElementsByTagName('gml:LineString')
                            for lineString in lineStrings:
                                varBBBA = QtGui.QTreeWidgetItem(varBBB, [lineString.nodeName])
                                self.getAttributes(lineString,varBBBA)
                                self.getNodes(lineString.childNodes,400,varBBBA)
                                        
       
    # Get the attribute information of a given node.
    def getAttributes(self, item, parent):
        if item.localName != None:
            keys = item.attributes.keys()
            if item.attributes.length >= 1:
                for key in keys:
                    QtGui.QTreeWidgetItem(parent, [str(key),item.attributes[key].value])                        
            datas = item.childNodes
            for data in datas:
                if (data.nodeName == '#text'):
                    if "\n" not in data.data:
                        QtGui.QTreeWidgetItem(parent, [data.parentNode.nodeName + data.nodeName,data.data])
                    

    # Exploring the whole of child nodes of a given node, getting attributes and data.
    def getNodes (self, item, level, parent):
        if level != 0:
            for node in item:
                if (node.nodeName == 'gml:Polygon') or (node.nodeName == 'gml:Point') or (node.nodeName == 'gml:LineString'):
                    break
                self.getAttributes(node,parent)
                self.getNodes(node.childNodes, level - 1, parent)
                
    
    # Offer a message in order to inform user about the need of upload an IndoorGML file.
    def showMessage(self):
        msg = QMessageBox()
        msg.setIconPixmap(QtGui.QPixmap(":/plugins/IndoorGMLReader/icon.png"));
        msg.setText("Please, load IndoorGML file")
        msg.setWindowTitle("WARNING - IndoorGML")
        msg.setDetailedText("Click on 'IndoorGML file' button.\nSelect the IndoorGML file.")
        msg.exec_()
    
    
    # Get feature information, through tree item click, on the QText element.
    def displayXlinkInformation(self, item, column):
       
        if (item.text(0) == 'xlink:href'):
            file = self.dlg.gmlLine.displayText()
            gml = minidom.parse(file)
            extendedNodes = gml.getElementsByTagName('Transition')
            for node in extendedNodes:
                if ((node.attributes['gml:id'].value == item.text(1)) or ('#' + node.attributes['gml:id'].value == item.text(1))):
                    self.createSecTree(node)
                    break
            extendedNodes = gml.getElementsByTagName('CellSpace')
            for node in extendedNodes:
                if ((node.attributes['gml:id'].value == item.text(1)) or ('#' + node.attributes['gml:id'].value == item.text(1))):
                    self.createSecTree(node)
                    break  
            extendedNodes = gml.getElementsByTagName('State')
            for node in extendedNodes:
                if ((node.attributes['gml:id'].value == item.text(1)) or ('#' + node.attributes['gml:id'].value == item.text(1))):
                    self.createSecTree(node)
                    break  
        else:
            self.dlg.secTree.clear()


    # The secondary tree is created. It will show infomration of xlink:href linked features.
    def createSecTree (self, node):
        secTree = self.dlg.secTree
        header = QtGui.QTreeWidgetItem(["xlink:href Feature Information","Value"])
        secTree.setHeaderItem(header)
        secTree.header().setResizeMode(QtGui.QHeaderView.ResizeToContents)
        secTree.header().setStretchLastSection(False)
        self.dlg.secTree.clear()
        root = QtGui.QTreeWidgetItem(secTree, [node.nodeName])
        varA = QtGui.QTreeWidgetItem(root, ['Properties'])
        self.getAttributes(node,varA)
        self.getNodes(node.childNodes,400,varA)
        if node.nodeName == 'State':
            points = node.getElementsByTagName('gml:Point')
            for point in points:
                varB = QtGui.QTreeWidgetItem(root, [point.nodeName])
                self.getAttributes(point,varB)
                self.getNodes(point.childNodes,400,varB)
        if node.nodeName == 'CellSpace':
            polygons = node.getElementsByTagName('gml:Polygon')
            for polygon in polygons:
                varB = QtGui.QTreeWidgetItem(root, [polygon.nodeName])
                self.getAttributes(polygon,varB)
                self.getNodes(polygon.childNodes,400,varB)
        if node.nodeName == 'Transition':
            lineStrings = node.getElementsByTagName('gml:LineString')
            for lineString in lineStrings:
                varB = QtGui.QTreeWidgetItem(root, [lineString.nodeName])
                self.getAttributes(lineString,varB)
                self.getNodes(lineString.childNodes,400,varB)
                     
       
    def printCellSpaces (self,gml):    
        # We get the filename in order to give name the layer.
        fileName = os.path.basename(os.path.normpath(self.dlg.gmlLine.displayText()))
        layerName = '(CellSpaces) - ' + fileName
        
        # In order to avoid repeated layers.
        layers = self.iface.legendInterface().layers()
        layerExists = 0
        for layer in layers:
            if layer.name() == layerName:
                layerExists = 1
        
        polygons = gml.getElementsByTagName('gml:Polygon')
        
        if ((polygons.length >= 1) and (layerExists == 0)):
        
            layer =  QgsVectorLayer('Polygon', layerName, "memory")
            pr = layer.dataProvider() 
            
            for polygon in polygons:
                positions = polygon.getElementsByTagName('gml:pos')
                points = []
                for position in positions:
                    stringPosition = position.childNodes[0].data
                    splitedPosition = stringPosition.split(' ')
                    point = QgsPoint(float(splitedPosition[0]),float(splitedPosition[1]))
                    points.append(point)
               
                pol = QgsFeature()
                pol.setGeometry(QgsGeometry.fromPolygon([points]))
                pr.addFeatures([pol])
                layer.updateExtents()
                               
            QgsMapLayerRegistry.instance().addMapLayers([layer])
            
            # Setting the zoom scale to this layer.
            canvas = qgis.utils.iface.mapCanvas()
            canvas.setExtent(layer.extent())
            
            
    # Creating States layer.        
    def printStates (self,gml):
        # We get the filename in order to give name the layer.
        fileName = os.path.basename(os.path.normpath(self.dlg.gmlLine.displayText()))
        layerName = '(States) - ' + fileName
        
        # In order to avoid repeated layers.
        layers = self.iface.legendInterface().layers()
        layerExists = 0
        for layer in layers:
            if layer.name() == layerName:
                layerExists = 1
        
        states = gml.getElementsByTagName('State')
        
        if ((states.length >= 1) and (layerExists == 0)):

            layer = QgsVectorLayer('Point', '(States) - ' + fileName , "memory")
            pr = layer.dataProvider() 
            
            for state in states:
                pos = state.getElementsByTagName('gml:pos')
                stringPosition = pos[0].childNodes[0].data
                splitedPosition = stringPosition.split(' ')
                pt = QgsFeature()
                point1 = QgsPoint(float(splitedPosition[0]),float(splitedPosition[1]))
                pt.setGeometry(QgsGeometry.fromPoint(point1))
                pr.addFeatures([pt])
                layer.updateExtents()      
                
            QgsMapLayerRegistry.instance().addMapLayers([layer])
            
            # Setting the zoom scale to this layer.
            canvas = qgis.utils.iface.mapCanvas()
            canvas.setExtent(layer.extent())
                       
    
    def printTransitions (self,gml): 
        # We get the filename in order to give name the layer.
        fileName = os.path.basename(os.path.normpath(self.dlg.gmlLine.displayText()))
        layerName = '(Transitions) - ' + fileName
        
        # In order to avoid repeated layers.
        layers = self.iface.legendInterface().layers()
        layerExists = 0
        for layer in layers:
            if layer.name() == layerName:
                layerExists = 1
        
        transitions = gml.getElementsByTagName('Transition')
        
        if ((transitions.length >= 1) and (layerExists == 0)):
        
            layer =  QgsVectorLayer('LineString', '(Transitions) - ' + fileName, "memory")
            pr = layer.dataProvider() 
            
            for transition in transitions:
                positions = transition.getElementsByTagName('gml:pos')
                points = []
                for position in positions:
                    stringPosition = position.childNodes[0].data
                    splitedPosition = stringPosition.split(' ')
                    point = QgsPoint(float(splitedPosition[0]),float(splitedPosition[1]))
                    points.append(point)
                    
                l = len(points)
                i = 0
                
                for point in points:
                    if l >= 2:
                        line = QgsFeature()
                        line.setGeometry(QgsGeometry.fromPolyline([points[i],points[i + 1]]))
                        pr.addFeatures([line])
                        layer.updateExtents()
                        i = i + 1 
                        l = l - 1
                
            QgsMapLayerRegistry.instance().addMapLayers([layer])
            
            # Setting the zoom scale to this layer.
            canvas = qgis.utils.iface.mapCanvas()
            canvas.setExtent(layer.extent())
            
            
    def run(self):
        """Run method that performs all the real work"""
        # show the dialog
        self.dlg.show()
        self.dlg.gmlLine.clear()
        self.dlg.tree.clear()
        self.dlg.secTree.clear()
        # Run the dialog event loop
        result = self.dlg.exec_()

